/*	Copyright (C) 2003-2013 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "oplab/oplab.pmh"

#define FE_PPO_DEBUG		FALSE

using namespace fe;

void PuppetOp::initialize(void)
{
	catalog<String>("icon")="FE_alpha";

	//* TODO
	catalogRemove("IdAttr");

	catalog<bool>("dampJoints")=false;
	catalog<String>("dampJoints","label")="Damping";
	catalog<String>("dampJoints","page")="Edit";
	catalog<bool>("dampJoints","joined")=true;

	catalog<Real>("dampRatio")=1.0;
	catalog<String>("dampRatio","label")="Envelope";
	catalog<String>("dampRatio","page")="Edit";

	//* pose is alternative reference
	catalog<bool>("poseJoints")=false;
	catalog<String>("poseJoints","label")="Repose";
	catalog<String>("poseJoints","page")="Edit";
	catalog<bool>("poseJoints","joined")=true;

	//* how much the pose changes the reference
	catalog<Real>("poseRatio")=1.0;
	catalog<String>("poseRatio","label")="Envelope";
	catalog<String>("poseRatio","page")="Edit";

	catalog<bool>("offsetJoints")=false;
	catalog<String>("offsetJoints","label")="Offset";
	catalog<String>("offsetJoints","page")="Edit";
	catalog<bool>("offsetJoints","joined")=true;

	catalog<Real>("offsetRatio")=1.0;
	catalog<String>("offsetRatio","label")="Envelope";
	catalog<String>("offsetRatio","page")="Edit";

	catalog<bool>("drawNames")=false;
	catalog<String>("drawNames","label")="Draw Names";
	catalog<String>("drawNames","page")="Display";

	catalog< sp<Component> >("Brush");
	catalog<String>("Brush","implementation")="DrawI";
	catalog<String>("Brush","prompt")="prompt here";

	catalog< sp<Component> >("Input Joints");

	catalog< sp<Component> >("Pose Joints");
	catalog<bool>("Pose Joints","optional")=true;

	catalog< sp<Component> >("Offset Joints");
	catalog<bool>("Offset Joints","optional")=true;

	m_spWireframe=new DrawMode();
	m_spWireframe->setDrawStyle(DrawMode::e_wireframe);
//	m_spWireframe->setPointSize(3.0);
//	m_spWireframe->setLineWidth(2.0);

	m_spSolid=new DrawMode();
	m_spSolid->copy(m_spWireframe);
	m_spSolid->setDrawStyle(DrawMode::e_solid);
	m_spSolid->setLit(FALSE);
	m_spSolid->setBackfaceCulling(FALSE);
}

void PuppetOp::handle(Record& a_rSignal)
{
#if FE_PPO_DEBUG
	feLog("PuppetOp::handle\n");
#endif

	sp<SurfaceAccessibleI> spInputAccessible;
	if(!access(spInputAccessible,"Input Joints")) return;

	sp<SurfaceAccessorI> spInputName;
	if(!access(spInputName,spInputAccessible,e_primitive,"name")) return;

	sp<SurfaceAccessorI> spInputParent;
	if(!access(spInputParent,spInputAccessible,e_primitive,"parent")) return;

	sp<SurfaceAccessorI> spInputColor;
	if(!access(spInputColor,spInputAccessible,e_point,"Cd")) return;

	const U32 primitiveCount=spInputName->count();

	sp<DrawI> spDrawBrush;
	accessBrush(spDrawBrush,a_rSignal,e_quiet);

	sp<DrawI> spDrawOverlay;
	accessBrushOverlay(spDrawOverlay,a_rSignal,e_quiet);

#if FE_PPO_DEBUG
	feLog("PuppetOp::handle brush %d brushed %d\n",
			spDrawBrush.isValid(),m_brushed);
#endif

	if(spDrawBrush.isValid())
	{
		sp<SurfaceAccessorI> spOutputVertices;
		if(m_spOutputAccessible.isNull() ||
				!access(spOutputVertices,m_spOutputAccessible,
				e_primitive,e_vertices)) return;

		spDrawBrush->pushDrawMode(m_spWireframe);
		spDrawOverlay->pushDrawMode(m_spSolid);

		sp<ViewI> spView=spDrawOverlay->view();

		m_brushed=TRUE;

		const Color black(0.0,0.0,0.0);
		const Color white(1.0,1.0,0.8);
		const Color grey(0.8,0.8,1.0);
		const Color red(0.8,0.0,0.0);
		const Color green(0.0,0.5,0.0);
		const Color blue(0.0,0.0,1.0);
		const Color cyan(0.0,1.0,1.0);
		const Color magenta(0.8,0.0,0.8);
		const Color yellow(1.0,1.0,0.0);
		const Color sparkle(1.0,1.0,0.6);
		const Color orange(1.0,0.5,0.0);
		const Color blank(0.0,0.0,0.0,0.0);

		m_event.bind(windowEvent(a_rSignal));

#if FE_PPO_DEBUG
		feLog("%s\n",c_print(m_event));
#endif
		if(m_event.isMousePress(WindowEvent::e_itemLeft))
		{
			const I32 mouseX=m_event.mouseX();
			const I32 mouseY=m_event.mouseY();

			feLog("press %d %d\n",mouseX,mouseY);
		}

		SpatialVector line[3];

		for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
		{
			const String parentName=spInputParent->string(primitiveIndex);

			const BWORD hasParent=(m_animMap.find(parentName)!=m_animMap.end());
			if(!hasParent)
			{
				continue;
			}

			const I32 subCount=spOutputVertices->subCount(primitiveIndex);
			if(subCount<3)
			{
				continue;
			}

			const String name=spInputName->string(primitiveIndex);

			const I32 pointIndex0=
					spOutputVertices->integer(primitiveIndex,0);
			const I32 pointIndex1=
					spOutputVertices->integer(primitiveIndex,1);
			const I32 pointIndex2=
					spOutputVertices->integer(primitiveIndex,2);

			const Color color0(spInputColor->spatialVector(pointIndex0));
			const Color color1(spInputColor->spatialVector(pointIndex1));
			const Color color2(spInputColor->spatialVector(pointIndex2));

			const SpatialVector point0=
					spOutputVertices->spatialVector(primitiveIndex,0);
			const SpatialVector point1=
					spOutputVertices->spatialVector(primitiveIndex,1);
			const SpatialVector point2=
					spOutputVertices->spatialVector(primitiveIndex,2);

			const SpatialVector mod1=2.0*point0-point1;
			const SpatialVector mod2=2.0*point0-point2;

//			const SpatialTransform& xform=m_animMap[name];
			const SpatialTransform& xformParent=m_animMap[parentName];

			const SpatialVector& parentPoint0=xformParent.translation();

			line[0]=parentPoint0;

			line[1]=point1;
			spDrawBrush->drawLines(line,NULL,2,DrawI::e_strip,false,&color1);

			line[1]=point2;
			spDrawBrush->drawLines(line,NULL,2,DrawI::e_strip,false,&color2);

			line[1]=mod1;
			line[2]=mod2;
			spDrawBrush->drawLines(line,NULL,3,DrawI::e_strip,false,&color0);

			line[1]=mod2;
			line[2]=point1;
			spDrawBrush->drawLines(line,NULL,3,DrawI::e_strip,false,&color0);

			line[0]=mod1;
			line[1]=point2;
			spDrawBrush->drawLines(line,NULL,2,DrawI::e_strip,false,&color0);

			if(catalog<bool>("drawNames"))
			{
				const U32 margin=1;
				const SpatialVector midpoint=0.5*(parentPoint0+point0);
				const Vector2i projected=
						spView->project(midpoint,ViewMode::e_perspective);
				drawLabel(spDrawOverlay,projected,name,
						FALSE,margin,white,NULL,&blue);
			}
		}

		spDrawOverlay->popDrawMode();
		spDrawBrush->popDrawMode();

		return;
	}

	const Real dampRatio=
			catalog<bool>("dampJoints")? catalog<Real>("dampRatio"): 0.0;
	const Real poseRatio=
			catalog<bool>("poseJoints")? catalog<Real>("poseRatio"): 0.0;
	const Real offsetRatio=
			catalog<bool>("offsetJoints")? catalog<Real>("offsetRatio"): 0.0;

	if(!accessOutput(m_spOutputAccessible,a_rSignal)) return;

	sp<SurfaceAccessorI> spOutputVertices;
	if(!access(spOutputVertices,m_spOutputAccessible,
			e_primitive,e_vertices)) return;

	sp<SurfaceAccessorI> spOutputPoint;
	if(!access(spOutputPoint,m_spOutputAccessible,e_point,e_position)) return;

	sp<SurfaceAccessorI> spOutputAnimX;
	if(!access(spOutputAnimX,m_spOutputAccessible,e_primitive,"animX")) return;

	sp<SurfaceAccessorI> spOutputAnimY;
	if(!access(spOutputAnimY,m_spOutputAccessible,e_primitive,"animY")) return;

	sp<SurfaceAccessorI> spOutputAnimZ;
	if(!access(spOutputAnimZ,m_spOutputAccessible,e_primitive,"animZ")) return;

	sp<SurfaceAccessorI> spOutputAnimT;
	if(!access(spOutputAnimT,m_spOutputAccessible,e_primitive,"animT")) return;

	sp<SurfaceAccessorI> spInputRefX;
	sp<SurfaceAccessorI> spInputRefY;
	sp<SurfaceAccessorI> spInputRefZ;
	sp<SurfaceAccessorI> spInputRefT;
	if(dampRatio>0.0 || offsetRatio>0.0)
	{
		if(!access(spInputRefX,spInputAccessible,e_primitive,"refX")) return;

		if(!access(spInputRefY,spInputAccessible,e_primitive,"refY")) return;

		if(!access(spInputRefZ,spInputAccessible,e_primitive,"refZ")) return;

		if(!access(spInputRefT,spInputAccessible,e_primitive,"refT")) return;
	}

	sp<SurfaceAccessibleI> spPoseAccessible;
	sp<SurfaceAccessorI> spPoseAnimX;
	sp<SurfaceAccessorI> spPoseAnimY;
	sp<SurfaceAccessorI> spPoseAnimZ;
	sp<SurfaceAccessorI> spPoseAnimT;
	if(poseRatio>0.0)
	{
		access(spPoseAccessible,"Pose Joints",e_quiet);

		if(spPoseAccessible.isValid())
		{
			if(!access(spPoseAnimX,spPoseAccessible,
					e_primitive,"animX")) return;

			if(!access(spPoseAnimY,spPoseAccessible,
					e_primitive,"animY")) return;

			if(!access(spPoseAnimZ,spPoseAccessible,
					e_primitive,"animZ")) return;

			if(!access(spPoseAnimT,spPoseAccessible,
					e_primitive,"animT")) return;
		}
	}

	sp<SurfaceAccessibleI> spOffsetAccessible;
	sp<SurfaceAccessorI> spOffsetAnimX;
	sp<SurfaceAccessorI> spOffsetAnimY;
	sp<SurfaceAccessorI> spOffsetAnimZ;
	sp<SurfaceAccessorI> spOffsetAnimT;
	if(offsetRatio>0.0)
	{
		access(spOffsetAccessible,"Offset Joints",e_quiet);

		if(spOffsetAccessible.isValid())
		{
			if(!access(spOffsetAnimX,spOffsetAccessible,
					e_primitive,"animX")) return;

			if(!access(spOffsetAnimY,spOffsetAccessible,
					e_primitive,"animY")) return;

			if(!access(spOffsetAnimZ,spOffsetAccessible,
					e_primitive,"animZ")) return;

			if(!access(spOffsetAnimT,spOffsetAccessible,
					e_primitive,"animT")) return;
		}
	}

	m_animMap.clear();

	for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
	{
		SpatialTransform xformLocalRef;
		SpatialQuaternion quatLocalRef;
		if(spInputRefX.isValid())
		{
			xformLocalRef.column(0)=spInputRefX->spatialVector(primitiveIndex);
			xformLocalRef.column(1)=spInputRefY->spatialVector(primitiveIndex);
			xformLocalRef.column(2)=spInputRefZ->spatialVector(primitiveIndex);
			xformLocalRef.column(3)=spInputRefT->spatialVector(primitiveIndex);

			quatLocalRef=xformLocalRef;
		}

		SpatialTransform xformLocal;
		xformLocal.column(0)=spOutputAnimX->spatialVector(primitiveIndex);
		xformLocal.column(1)=spOutputAnimY->spatialVector(primitiveIndex);
		xformLocal.column(2)=spOutputAnimZ->spatialVector(primitiveIndex);
		xformLocal.column(3)=spOutputAnimT->spatialVector(primitiveIndex);

		SpatialQuaternion quatPose(0.0,0.0,0.0,1.0);
		if(spPoseAccessible.isValid())
		{
			SpatialTransform xformPose;
			xformPose.column(0)=spPoseAnimX->spatialVector(primitiveIndex);
			xformPose.column(1)=spPoseAnimY->spatialVector(primitiveIndex);
			xformPose.column(2)=spPoseAnimZ->spatialVector(primitiveIndex);
			xformPose.column(3)=spPoseAnimT->spatialVector(primitiveIndex);

			SpatialTransform invLocalRef;
			invert(invLocalRef,xformLocalRef);

			quatPose=xformPose*invLocalRef;

			Real radians;
			SpatialVector axis;
			quatPose.computeAngleAxis(radians,axis);
			const SpatialQuaternion quatPartial(poseRatio*radians,axis);

			quatLocalRef=quatPartial*quatLocalRef;
		}

		SpatialQuaternion quatOffset(0.0,0.0,0.0,1.0);
		if(spOffsetAccessible.isValid())
		{
			SpatialTransform xformOffset;
			xformOffset.column(0)=spOffsetAnimX->spatialVector(primitiveIndex);
			xformOffset.column(1)=spOffsetAnimY->spatialVector(primitiveIndex);
			xformOffset.column(2)=spOffsetAnimZ->spatialVector(primitiveIndex);
			xformOffset.column(3)=spOffsetAnimT->spatialVector(primitiveIndex);

			SpatialTransform invLocalRef;
			invert(invLocalRef,xformLocalRef);

			quatOffset=xformOffset*invLocalRef;
		}

		const SpatialVector localDisplace=xformLocal.translation();

		SpatialQuaternion quatLocal=xformLocal;

		//* TODO blend local offset with a reference offset

		SpatialQuaternion quatBlended;
		if(dampRatio>0.0)
		{
			Real radians;
			SpatialVector axis;
			quatLocal.computeAngleAxis(radians,axis);
			const SpatialQuaternion quatPartial(
					(1.0-dampRatio)*radians,axis);

			Real radiansRef;
			SpatialVector axisRef;
			quatLocalRef.computeAngleAxis(radiansRef,axisRef);
			const SpatialQuaternion quatPartialRef(
					dampRatio*radiansRef,axisRef);

			quatBlended=quatPartial*quatPartialRef;
		}
		else
		{
			quatBlended=quatLocal;
		}

		if(offsetRatio>0.0)
		{
			Real radiansOff;
			SpatialVector axisOff;
			quatOffset.computeAngleAxis(radiansOff,axisOff);
			const SpatialQuaternion quatPartialOff(
					offsetRatio*radiansOff,axisOff);

			quatBlended=quatBlended*quatPartialOff;
		}

		SpatialVector blendedDisplace;
		rotateVector(quatBlended*(-quatLocal),localDisplace,blendedDisplace);

		SpatialTransform xformBlended(quatBlended);
		xformBlended.translation()=blendedDisplace;

		spOutputAnimX->set(primitiveIndex,xformBlended.column(0));
		spOutputAnimY->set(primitiveIndex,xformBlended.column(1));
		spOutputAnimZ->set(primitiveIndex,xformBlended.column(2));
		spOutputAnimT->set(primitiveIndex,xformBlended.column(3));

		//* update visible triangle

		const String name=spInputName->string(primitiveIndex);
		const String parentName=spInputParent->string(primitiveIndex);

		SpatialTransform xformParent;

		const BWORD hasParent=(m_animMap.find(parentName)==m_animMap.end());
		if(hasParent)
		{
			setIdentity(xformParent);
		}
		else
		{
			xformParent=m_animMap[parentName];
		}

		const SpatialTransform xformWorld=xformBlended*xformParent;
		m_animMap[name]=xformWorld;

		//* TODO param
		const Real triScale=1.0;

		const I32 pointIndex0=spOutputVertices->integer(primitiveIndex,0);
		const I32 pointIndex1=spOutputVertices->integer(primitiveIndex,1);
		const I32 pointIndex2=spOutputVertices->integer(primitiveIndex,2);

		const SpatialVector point=xformWorld.translation();
		spOutputPoint->set(pointIndex0,point);
		spOutputPoint->set(pointIndex1,point+triScale*xformWorld.column(0));
		spOutputPoint->set(pointIndex2,point+triScale*xformWorld.column(2));
	}

#if FE_PPO_DEBUG
	feLog("PuppetOp::handle done\n");
#endif
}
