/*	Copyright (C) 2003-2013 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __platform_error_h__
#define __platform_error_h__

#if FE_OS==FE_WIN32 || FE_OS==FE_WIN64

#define FE_ERRNO	WSAGetLastError()
#define FE_BEGIN_SIGNAL_REDO_LOOP();		do{
#define FE_END_SIGNAL_REDO_LOOP(if_error)	} while ((if_error) && (FE_ERRNO == WSAEINTR))
#define FE_UNDEFINED_ERROR		0xFFF0
#define	FE_EWOULDBLOCK			WSAEWOULDBLOCK
#define	FE_EBADF				WSAEBADF
#define	FE_ENOTSOCK				WSAENOTSOCK
#define	FE_EINTR				WSAEINTR
#define	FE_ENODEV				FE_UNDEFINED_ERROR+0
#define	FE_EOPNOTSUPP			WSAEOPNOTSUPP
#define	FE_EPROTO				FE_UNDEFINED_ERROR+1
#define	FE_ENOMEM				FE_UNDEFINED_ERROR+2
#define	FE_ENOSR				FE_UNDEFINED_ERROR+3
#define	FE_EACCES				WSAEACCES
#define	FE_EADDRINUSE			WSAEADDRINUSE
#define	FE_EADDRNOTAVAIL		WSAEADDRNOTAVAIL
#define	FE_EAFNOSUPPORT			WSAEAFNOSUPPORT
#define	FE_ENOENT				FE_UNDEFINED_ERROR+4
#define	FE_ECONNREFUSED			WSAECONNREFUSED
#define	FE_EINPROGRESS			WSAEINPROGRESS
#define	FE_EINVAL				WSAEINVAL
#define	FE_EIO					FE_UNDEFINED_ERROR+5
#define	FE_ENXIO				FE_UNDEFINED_ERROR+6
#define	FE_ECONNRESET			WSAECONNRESET
#define	FE_EISCONN				WSAEISCONN
#define	FE_ENETUNREACH			WSAENETUNREACH
#define	FE_ETIMEDOUT			WSAETIMEDOUT
#define	FE_EMFILE				WSAEMFILE
#define	FE_EPROTONOSUPPORT		WSAEPROTONOSUPPORT
#define	FE_ESTALE				WSAESTALE
#define	FE_EMSGSIZE				WSAEMSGSIZE
#define FE_EPIPE				FE_UNDEFINED_ERROR+7
#define	FE_ENOTINIT				WSANOTINITIALISED
#endif

//* GENERAL LINUX
#if FE_OS==FE_LINUX || FE_OS==FE_OSX

#define GLAPIENTRY
//#define GLAPI

#define FE_ERRNO	errno
#define FE_BEGIN_SIGNAL_REDO_LOOP();		do{
#define FE_END_SIGNAL_REDO_LOOP(if_error)	} while ((if_error) && (FE_ERRNO == EINTR))
#define	FE_EWOULDBLOCK			EWOULDBLOCK
#define	FE_EBADF				EBADF
#define	FE_ENOTSOCK				ENOTSOCK
#define	FE_EINTR				EINTR
#define	FE_ENODEV				ENODEV
#define	FE_EOPNOTSUPP			EOPNOTSUPP
#define	FE_EPROTO				EPROTO
#define	FE_ENOMEM				ENOMEM
#define	FE_ENOSR				ENOSR
#define	FE_EACCES				EACCES
#define	FE_EADDRINUSE			EADDRINUSE
#define	FE_EADDRNOTAVAIL		EADDRNOTAVAIL
#define	FE_EAFNOSUPPORT			EAFNOSUPPORT
#define	FE_ENOENT				ENOENT
#define	FE_ECONNREFUSED			ECONNREFUSED
#define	FE_EINPROGRESS			EINPROGRESS
#define	FE_EINVAL				EINVAL
#define	FE_EIO					EIO
#define	FE_ENXIO				ENXIO
#define	FE_ECONNRESET			ECONNRESET
#define	FE_EISCONN				EISCONN
#define	FE_ENETUNREACH			ENETUNREACH
#define	FE_ETIMEDOUT			ETIMEDOUT
#define	FE_EMFILE				EMFILE
#define	FE_EPROTONOSUPPORT		EPROTONOSUPPORT
#define	FE_ESTALE				ESTALE
#define	FE_EMSGSIZE				EMSGSIZE
#define FE_EPIPE				EPIPE
#define	FE_ENOTINIT				FE_UNDEFINED_ERROR
#endif


//* SYSTEM DEPENDENT FUNCTION NAMES
#if FE_OS==FE_WIN32 || FE_OS==FE_WIN64
#define FE_VSNPRINTF _vsnprintf
#define FE_SNPRINTF _snprintf
#endif

#if FE_OS==FE_LINUX || FE_OS==FE_OSX
#define FE_VSNPRINTF vsnprintf
#define FE_SNPRINTF snprintf
#endif

#endif /* __platform_error_h__ */
